/*
 * Copyright 2007 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.octo.mtg.plugin.tools;

import org.apache.maven.model.Plugin;
import org.apache.maven.model.PluginManagement;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.project.MavenProject;
import org.codehaus.plexus.logging.AbstractLogEnabled;
import org.codehaus.plexus.util.IOUtil;
import org.codehaus.plexus.util.cli.CommandLineException;
import org.codehaus.plexus.util.cli.Commandline;
import org.codehaus.plexus.util.xml.Xpp3Dom;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.text.DateFormat;
import java.util.Calendar;

/**
 * @author <a href="mailto:aheritier@gmail.com">Arnaud HERITIER</a>
 * @version $Id: DefaultGrailsServices.java 468 2008-05-26 15:50:09Z ahe $
 * @plexus.component role="com.octo.mtg.plugin.tools.GrailsServices"
 * @since 0.1
 */
public class DefaultGrailsServices extends AbstractLogEnabled implements GrailsServices {

    /**
     * @plexus.requirement
     */
    private PomServices pomServices;

    /**
     * @plexus.requirement
     */
    private MojoServices mojoServices;

    private File b;

    /**
     * OutputStream to write the content of stdout.
     */
    private OutputStream infoOutputStream = new OutputStream() {
        StringBuffer buffer = new StringBuffer();

        public void write(int b) throws IOException {
            if (b == '\n') {
                getLogger().info(buffer.toString());
                buffer.setLength(0);
            } else {
                buffer.append((char) b);
            }
        }
    };

    /**
     * OutputStream to write the content of stderr.
     */
    private OutputStream warnOutputStream = new OutputStream() {
        StringBuffer buffer = new StringBuffer();

        public void write(int b) throws IOException {
            if (b == '\n') {
                getLogger().warn(buffer.toString());
                buffer.setLength(0);
            } else {
                buffer.append((char) b);
            }
        }
    };

    private File getBasedir() {
        if (b != null) {
            return b;
        }

        throw new RuntimeException("The basedir has to be set before any of the service methods are invoked.");
    }

    // -----------------------------------------------------------------------
    // GrailsServices Implementation
    // -----------------------------------------------------------------------

    public void setBasedir(File basedir) {
        this.b = basedir;
    }

    public MavenProject createPOM(String groupId, GrailsProject grailsProjectDescriptor, String mtgGroupId,
                                  String grailsPluginArtifactId, String mtgVersion) {
        return createPOM(groupId, grailsProjectDescriptor, mtgGroupId, grailsPluginArtifactId, mtgVersion, false);
    }

    public MavenProject createPOM(String groupId,
                                  GrailsProject grailsProjectDescriptor, String mtgGroupId,
                                  String grailsPluginArtifactId, String mtgVersion,
                                  boolean addEclipseSettings) {

        MavenProject pom = new MavenProject();
        if (pom.getBuild().getPluginManagement() == null)
            pom.getBuild().setPluginManagement(new PluginManagement());
        PluginManagement pluginMgt = pom.getPluginManagement();

        // Those four properties are needed.
        pom.setModelVersion("4.0.0");
        pom.setPackaging("grails-app");
        // Specific for GRAILS
        pom.getModel().getProperties().setProperty("grailsHome", "${env.GRAILS_HOME}");
        pom.getModel().getProperties().setProperty("grailsVersion", grailsProjectDescriptor.getAppGrailsVersion());
        // Add our own plugin
        Plugin grailsPlugin = new Plugin();
        grailsPlugin.setGroupId(mtgGroupId);
        grailsPlugin.setArtifactId(grailsPluginArtifactId);
        grailsPlugin.setVersion(mtgVersion);
        grailsPlugin.setExtensions(true);
        pom.addPlugin(grailsPlugin);
        // Add compiler plugin settings
        Plugin compilerPlugin = new Plugin();
        compilerPlugin.setGroupId("org.apache.maven.plugins");
        compilerPlugin.setArtifactId("maven-compiler-plugin");
        Xpp3Dom compilerConfig = new Xpp3Dom("configuration");
        Xpp3Dom source = new Xpp3Dom("source");
        source.setValue("1.5");
        compilerConfig.addChild(source);
        Xpp3Dom target = new Xpp3Dom("target");
        target.setValue("1.5");
        compilerConfig.addChild(target);
        compilerPlugin.setConfiguration(compilerConfig);
        pom.addPlugin(compilerPlugin);
        // Add eclipse plugin settings
        if (addEclipseSettings) {
            Plugin warPlugin = new Plugin();
            warPlugin.setGroupId("org.apache.maven.plugins");
            warPlugin.setArtifactId("maven-war-plugin");
            Xpp3Dom warConfig = new Xpp3Dom("configuration");
            Xpp3Dom warSourceDirectory = new Xpp3Dom("warSourceDirectory");
            warSourceDirectory.setValue("web-app");
            warConfig.addChild(warSourceDirectory);
            warPlugin.setConfiguration(warConfig);
            pluginMgt.addPlugin(warPlugin);

            Plugin eclipsePlugin = new Plugin();
            eclipsePlugin.setGroupId("org.apache.maven.plugins");
            eclipsePlugin.setArtifactId("maven-eclipse-plugin");
            Xpp3Dom configuration = new Xpp3Dom("configuration");
            Xpp3Dom projectnatures = new Xpp3Dom("additionalProjectnatures");
            Xpp3Dom projectnature = new Xpp3Dom("projectnature");
            projectnature.setValue("org.codehaus.groovy.eclipse.groovyNature");
            projectnatures.addChild(projectnature);
            configuration.addChild(projectnatures);
            Xpp3Dom additionalBuildcommands = new Xpp3Dom(
                "additionalBuildcommands");
            Xpp3Dom buildcommand = new Xpp3Dom("buildcommand");
            buildcommand.setValue("org.codehaus.groovy.eclipse.groovyBuilder");
            additionalBuildcommands.addChild(buildcommand);
            configuration.addChild(additionalBuildcommands);
            // Xpp3Dom additionalProjectFacets = new Xpp3Dom(
            // "additionalProjectFacets");
            // Xpp3Dom jstWeb = new Xpp3Dom("jst.web");
            // jstWeb.setValue("2.5");
            // additionalProjectFacets.addChild(jstWeb);
            // configuration.addChild(additionalProjectFacets);
            Xpp3Dom packaging = new Xpp3Dom("packaging");
            packaging.setValue("war");
            configuration.addChild(packaging);

            eclipsePlugin.setConfiguration(configuration);
            pluginMgt.addPlugin(eclipsePlugin);
        }
        // Change the default output directory to generate classes
        pom.getModel().getBuild().setOutputDirectory("web-app/WEB-INF/classes");

        pom.setArtifactId(grailsProjectDescriptor.getAppName());
        pom.setName(grailsProjectDescriptor.getAppName());
        pom.setGroupId(groupId);
        pom.setVersion(grailsProjectDescriptor.getAppVersion());
        if (!grailsProjectDescriptor.getAppVersion().endsWith("SNAPSHOT")) {
            getLogger().warn("=====================================================================");
            getLogger().warn("If your project is currently in development, in accordance with maven ");
            getLogger().warn("standards, its version must be " + grailsProjectDescriptor.getAppVersion() + "-SNAPSHOT and not " + grailsProjectDescriptor.getAppVersion() + ".");
            getLogger().warn("Please, change your version in the application.properties descriptor");
            getLogger().warn("and regenerate your pom.");
            getLogger().warn("=====================================================================");
        }
        return pom;
    }

    public GrailsProject readProjectDescriptor() throws IOException {

        // Load existing Grails properties
        GrailsProject grailsProps = new GrailsProject();
        FileInputStream fis = null;
        try {
            fis = new FileInputStream(new File(getBasedir(), "application.properties"));
            grailsProps.load(fis);
            // We have to set the application version in grails settings for old
            // versions
            if (grailsProps.getAppVersion() == null) {
                grailsProps.setAppVersion(GrailsProject.DEFAULT_APP_VERSION);
                getLogger().warn("application.properties didn't contain an app.version property");
                getLogger().warn("Setting to default value '" + grailsProps.getAppVersion() + "'.");

                writeProjectDescriptor(getBasedir(), grailsProps);
            }
            return grailsProps;
        } finally {
            IOUtil.close(fis);
        }
    }

    public void writeProjectDescriptor(File projectDir, GrailsProject grailsProjectDescriptor)
        throws IOException {

        String description = "Grails Descriptor updated by grails-maven-plugin on " +
            DateFormat.getDateInstance().format(Calendar.getInstance().getTime());

        FileOutputStream fos = null;
        try {
            fos = new FileOutputStream(new File(projectDir, "application.properties"));
            grailsProjectDescriptor.store(fos, description);
        } finally {
            IOUtil.close(fos);
        }
    }

    /**
     * Launch a GRAILS commandLine with a target and no parameter.
     *
     * @param targetName the grails target to call
     * @throws CommandLineException If a problem occurs to call Grails.
     */
    public void launchGrails(String grailsHome, String env, String targetName) throws MojoExecutionException {
        launchGrails(grailsHome, env, targetName, null);
    }

    /**
     * Launch a GRAILS commandLine with a target and its parameters.
     *
     * @param targetName   the grails target to call
     * @param targetParams the target parameters
     * @throws CommandLineException If a problem occurs to call Grails.
     */
    public void launchGrails(String grailsHome, String env, String targetName, String[] targetParams)
        throws MojoExecutionException {
        Commandline cmd = new Commandline();
        try {
            cmd.addSystemEnvironment();
        } catch (Exception e1) {
            getLogger().warn("Unable to use system environment variables to launch Grails", e1);
        }

        getLogger().debug("Setting environment variable GRAILS_HOME to '" + grailsHome + "'.");
        cmd.addEnvironment("GRAILS_HOME", grailsHome);

        getLogger().debug("Working directory " + getBasedir() + "'.");
        cmd.setWorkingDirectory(getBasedir());

        cmd.setExecutable(grailsHome + "/bin/grails");

        // Setup grails env
        if (env != null) {
            // For default environments, we use the command line arg
            // as a workaround for [GRAILS-1658]
            if ("dev".equals(env)) {
                cmd.createArg().setValue("dev");
            } else if ("prod".equals(env)) {
                cmd.createArg().setValue("prod");
            } else if ("test".equals(env)) {
                cmd.createArg().setValue("test");
            } else {
                cmd.createArg().setValue("-Dgrails.env=" + env);
            }
        }

        cmd.createArg().setValue(targetName);
        if (targetParams != null) {
            for (int i = 0; i < targetParams.length; i++) {
                if (targetParams[i] != null)
                    cmd.createArg().setValue(targetParams[i]);
            }
        }

        try {
            getLogger().debug("Command line: " + cmd.toString());
            int returnCode = mojoServices.executeCommandLine(cmd, System.in, infoOutputStream, warnOutputStream);
            getLogger().debug("Grails ended with the return code: " + returnCode);
            if (returnCode != 0) {
                throw new MojoExecutionException("Grails ended with a non null return code: " + returnCode);
            }
        } catch (CommandLineException e) {
            throw new MojoExecutionException("Unable to execute grails", e);
        }
    }
}
